import numpy as np
from collections import Counter

def get_genomes(tree):
    genomes = {}
    for leaf in tree.iter_leaves():
        genomes[leaf.name] = leaf.genome
    return genomes

def collapse_genomes(genomes, normal_diploid_genome, num_chroms):
    for chrom in range(num_chroms):
        for leaf, genome in genomes.items():
            for allele in [0, 1]:
                genome[chrom][allele] = dict(Counter(genome[chrom][allele]))
                for i in set(normal_diploid_genome[chrom][allele]) - set(genome[chrom][allele].keys()):
                    genome[chrom][allele][i] = 0
    return genomes
                
def format_CN_profiles(tree, normal_diploid_genome, num_chroms, region_length, bin_len):
    # Consolidate leaf genomes
    genomes = get_genomes(tree)
    genome_counters = collapse_genomes(genomes, normal_diploid_genome, num_chroms)

    # Assign regions to bins
    bins, bin_coords = {}, {}
    regions_per_bin = np.floor(bin_len / region_length)
    for chrom in range(num_chroms):
        bins[chrom], bin_coords[chrom] = {}, {}
        bin_count = 0
        for i in normal_diploid_genome[chrom][0]:
            if i == 0:
                bins[chrom][bin_count] = [i]
            else:
                if i % regions_per_bin == 0:
                    bin_coords[chrom][bin_count] = [bins[chrom][bin_count][0]*region_length, (bins[chrom][bin_count][-1]+1)*region_length]
                    bin_count += 1
                    bins[chrom][bin_count] = [i]
                else:
                    bins[chrom][bin_count].append(i)
        bin_coords[chrom][bin_count] = [bins[chrom][bin_count][0]*region_length, (bins[chrom][bin_count][-1]+1)*region_length]

    # Find the average number of copies over the regions in each bin
    CN_profiles = {}
    for chrom in range(num_chroms):
        CN_profiles[chrom] = {}
        sorted_bins = list(bins[chrom].keys())
        sorted_bins.sort()
        for leaf, genome in genome_counters.items():
            CN_profiles[chrom][leaf] = [[], []]
            for b in sorted_bins:
                regions = bins[chrom][b]
                for allele in [0, 1]:
                    CN_profiles[chrom][leaf][allele].append(round(sum([genome[chrom][allele][i] for i in regions]) / len(regions)))
    
    return CN_profiles, bin_coords

def save_CN_profiles_leaves(tree, chrom_names, bins, region_length, filepath):
    f = open(filepath, 'w+')
    headers = ['CELL', 'chrom', 'start', 'end', 'CN states']
    f.write('\t'.join(headers) + '\n')

    leaves = [leaf for leaf in tree.iter_leaves()]

    # Writes to file in order of chroms, bins, cell
    for chrom in chrom_names:
        num_bins = len(bins[chrom]) - 1
        for i in range(num_bins):
            bin_start, bin_end = str(bins[chrom][i]*region_length), str(bins[chrom][i+1]*region_length)
            for leaf in leaves:
                CN_state = str(leaf.profile[chrom][0][i]) + ',' + str(leaf.profile[chrom][1][i])
                line = [leaf.name, chrom, bin_start, bin_end, CN_state]
                f.write('\t'.join(line) + '\n')
    f.close()

def save_CN_profiles_ancestors(tree, chrom_names, bins, region_length, filepath):
    f = open(filepath, 'w+')
    headers = ['CELL', 'chrom', 'start', 'end', 'CN states']
    f.write('\t'.join(headers) + '\n')

    ancestors = [node for node in tree.iter_preorder() if not node.is_leaf()]

    # Writes to file in order of chroms, bins, cell
    for chrom in chrom_names:
        num_bins = len(bins[chrom]) - 1
        for i in range(num_bins):
            bin_start, bin_end = str(bins[chrom][i]*region_length), str(bins[chrom][i+1]*region_length)
            for ancestor in ancestors:
                CN_state = str(ancestor.profile[chrom][0][i]) + ',' + str(ancestor.profile[chrom][1][i])
                line = [ancestor.name, chrom, bin_start, bin_end, CN_state]
                f.write('\t'.join(line) + '\n')
    f.close()
            



import os
import subprocess
import itertools
from collections import Counter
import pickle
from itertools import repeat
import multiprocessing
from pyfaidx import Fasta
from Bio import SeqIO

import numpy as np
from scipy.stats import beta, poisson
from scipy.optimize import newton_krylov, NoConvergence

from .sequence import *

def iter_by_chunk(iterable, chunksize):
    return itertools.zip_longest(*[iter(iterable)] * chunksize)

def init_pool_processes():
    np.random.seed()

# Estimating alpha/beta from point on lorenz curve
def get_alpha_beta(x0, y0):
    def F(P):
        Aa, Bb = P[0], P[1]
        X = beta.cdf(float(Aa)/(Aa+Bb), Aa, Bb) - x0
        Y = beta.cdf(float(Aa)/(Aa+Bb), Aa+1, Bb) - y0
        return [X, Y]
    guess = [10, 5]
    max_n = 1000
    n = 1
    while n < max_n:
        try:
            sol = newton_krylov(F, guess, method = 'lgmres', verbose = 0, rdiff = 0.1, maxiter=50)
            break
        except NoConvergence as e:
            guess = np.random.rand(2) * 10 + 0.1
        except ValueError as e:
            guess = np.random.rand(2) * 10 + 0.1
        n += 1
    if n == max_n:
        # Error has occurred here
        return [1.38, 1.38]
    else:
        return sol

def bezier_coef(points):
    n = len(points) - 1

    C = 4 * np.identity(n)
    np.fill_diagonal(C[1:], 1)
    np.fill_diagonal(C[:, 1:], 1)
    C[0, 0] = 2
    C[n - 1, n - 1] = 7
    C[n - 1, n - 2] = 2

    P = [2 * (2 * points[i] + points[i + 1]) for i in range(n)]
    P[0] = points[0] + 2 * points[1]
    P[n - 1] = 8 * points[n - 1] + points[n]

    A = np.linalg.solve(C, P)
    B = [0] * n
    for i in range(n - 1):
        B[i] = 2 * points[i + 1] - A[i + 1]
    B[n - 1] = (A[n - 1] + points[n]) / 2

    return A, B

def single_cubic_bezier(a, b, c, d):
    return lambda t: np.power(1 - t, 3) * a + 3 * np.power(1 - t, 2) * t * b + 3 * (1 - t) * np.power(t, 2) * c + np.power(t, 3) * d

def multiple_cubic_bezier(points):
    A, B = bezier_coef(points)
    return [single_cubic_bezier(points[i], A[i], B[i], points[i + 1]) for i in range(len(points) - 1)]
    
# Generates starting coverage points at set intervals 
def gen_start_interval(num_windows, interval, Aa, Bb):
    # Sample point every interval bin
    x = [i for i in range(0, num_windows, interval)]
    if x[-1] != num_windows-1:
        if num_windows - x[-1] >= interval/2:
            x.append(num_windows-1)
        else:
            x[-1] = num_windows-1
    # draw coverage from beta distribution
    y = list(np.random.beta(Aa, Bb, len(x)))
    points = [np.array(p) for p in zip(x, y)]
    return points

# Smooths points across bins with bezier curves
def gen_coverage(num_windows, interval, Aa, Bb):
    points = gen_start_interval(num_windows, interval, Aa, Bb)
    A, B = bezier_coef(points)
    curves = [single_cubic_bezier(points[i], A[i], B[i], points[i + 1]) for i in range(len(points) - 1)]

    new_points = []
    for i in range(len(points) - 1):
        f = curves[i]
        gaps = points[i+1][0] - points[i][0] + 1
        coords = [f(t) for t in np.linspace(0, 1, int(gaps))]
        if i == 0:
            new_points.append((round(coords[0][0]), coords[0][1]))
        new_points += [(round(i[0]), i[1]) for i in coords[1:]]
    
    new_points.sort(key=lambda x: x[0])
    # Multiply by 2 so that values are in range 0-2. Value of 1 indicates mean coverage.
    new_points = [2*max(min(i[1], 1), 0) for i in new_points]
    return new_points

# Draw readcounts for each bin
def draw_readcounts(num_windows, window_size, interval, Aa, Bb, coverage, read_len):
    avg_read = (coverage*window_size) / (2*read_len)
    cov_scales = gen_coverage(num_windows, interval, Aa, Bb)
    #exp_counts = [avg_read*x for x in cov_scales]
    #readcounts = [np.random.poisson(x) for x in exp_counts]
    readcounts = [poisson.rvs(avg_read*x) for x in cov_scales]
    return readcounts

# Generate reads across the genome for a given cell
def gen_reads_cell(cell, chrom_names, uniform_coverage, window_size, interval, Aa, Bb, coverage, read_len, seq_error, out_path):
    #print('Generating reads for:', cell.name)
    prefix = os.path.join(out_path, cell.name)
    cell_ref1 = prefix + '_allele0.fa'
    cell_ref2 = prefix + '_allele1.fa'
    counts = []
    for allele, cell_ref in enumerate([cell_ref1, cell_ref2]):
        init = False
        chrom_lens = read_fasta(cell_ref, chrom_lens_only = True)
        for chrom in chrom_names:
            if chrom in chrom_lens:
                num_windows = round(chrom_lens[chrom] / window_size)
                if not uniform_coverage:
                    readcounts = draw_readcounts(num_windows, window_size, interval, Aa, Bb, coverage, read_len)
                for w in range(num_windows):
                    start = w * window_size + 1
                    if w == num_windows - 1:
                        end = chrom_lens[chrom]
                    else:
                        end = (w+1) * window_size
                    
                    region_fa_path = cell_ref[:-3] + chrom + '-' + str(start) + '.fa'
                    with open(region_fa_path, 'w+') as f:
                        call = subprocess.run(['samtools', 'faidx', cell_ref, chrom + ':' + str(start) + '-' + str(end)], stdout=f)
                    
                    for record in SeqIO.parse(region_fa_path, 'fasta'):
                        total_N = record.seq.count('N')
                    total_bp = end - start + 1
                    N_ratio = total_N / total_bp
                    if uniform_coverage:
                        readcount = round(poisson.rvs((total_bp*coverage) / (2*read_len)) * (1-N_ratio))
                    else:
                        readcount = round(readcounts[w] * (1-N_ratio))

                    counts.append((allele, chrom, start, end, readcount))

                    if readcount > 0:
                        #seed = np.random.randint(1, 10000000)
                        proc = subprocess.run(['dwgsim', '-H', '-o', '1', '-N', str(readcount), '-1', str(read_len), '-2', str(read_len), '-e', str(seq_error), '-E', str(seq_error), region_fa_path, region_fa_path[:-3]], capture_output=True, text=True)
                        if not init:
                            os.rename(region_fa_path[:-3] + '.bwa.read1.fastq.gz', cell_ref[:-3] + '.read1.fastq.gz')
                            os.rename(region_fa_path[:-3] + '.bwa.read2.fastq.gz', cell_ref[:-3] + '.read2.fastq.gz')
                            init = True
                        else:
                            with open(cell_ref[:-3] + '.read1.fastq.gz', 'a') as f1, open(cell_ref[:-3] + '.read2.fastq.gz', 'a') as f2:
                                call = subprocess.run(['cat', region_fa_path[:-3] + '.bwa.read1.fastq.gz'], stdout=f1)
                                call = subprocess.run(['cat', region_fa_path[:-3] + '.bwa.read2.fastq.gz'], stdout=f2)
                            os.remove(region_fa_path[:-3] + '.bwa.read1.fastq.gz')
                            os.remove(region_fa_path[:-3] + '.bwa.read2.fastq.gz')
                        os.remove(region_fa_path[:-3] + '.mutations.txt')
                        os.remove(region_fa_path[:-3] + '.mutations.vcf')
                    os.remove(region_fa_path)
                    #os.remove(region_fa_path + '.fai')
        os.remove(cell_ref)
        os.remove(cell_ref + '.fai')
    with open(prefix + '.read1.fastq.gz', 'w+') as f1, open(prefix + '.read2.fastq.gz', 'w+') as f2:
        call = subprocess.run(['cat', prefix + '_allele0.read1.fastq.gz', prefix + '_allele1.read1.fastq.gz'], stdout=f1)
        call = subprocess.run(['cat', prefix + '_allele0.read2.fastq.gz', prefix + '_allele1.read2.fastq.gz'], stdout=f2)
    os.remove(prefix + '_allele0.read1.fastq.gz')
    os.remove(prefix + '_allele1.read1.fastq.gz')
    os.remove(prefix + '_allele0.read2.fastq.gz')
    os.remove(prefix + '_allele1.read2.fastq.gz')
    os.remove(prefix + '.pkl')

    with open(os.path.join(out_path, cell.name + '.readcounts.tsv'), 'w+') as f:
        for c in counts:
            f.write(f'{c[0]}\t{c[1]}\t{c[2]}\t{c[3]}\t{c[4]}\n')

# Generate reads for all cells
def gen_reads(ref1, ref2, num_regions, chrom_names, tree, uniform_coverage, x0, y0, region_length, interval, window_size, coverage, read_len, seq_error, out_path, num_processors):
    leaves = list(tree.iter_leaves())

    [Aa, Bb] = get_alpha_beta(x0, y0)
    if num_processors == 1:
        for cell in leaves:
            prefix = os.path.join(out_path, cell.name)
            for allele, cur_ref in enumerate([ref1, ref2]):
                cell_ref, chrom_lens = build_cell_ref(prefix + '.pkl', cur_ref, chrom_names, num_regions, region_length, allele, prefix)
            gen_reads_cell(cell, chrom_names, uniform_coverage, window_size, interval, Aa, Bb, coverage, read_len, seq_error, out_path)
    else:
        pool = multiprocessing.Pool(processes=num_processors)
        for chunk in iter_by_chunk(tree.iter_leaves(), num_processors):
            cells = list(chunk)
            while None in cells:
                cells.remove(None)
            for cell in cells:
                prefix = os.path.join(out_path, cell.name)
                for allele, cur_ref in enumerate([ref1, ref2]):
                    cell_ref, chrom_lens = build_cell_ref(prefix + '.pkl', cur_ref, chrom_names, num_regions, region_length, allele, prefix)
            with multiprocessing.Pool(processes=num_processors) as pool:
                pool.starmap(gen_reads_cell, zip(cells, repeat(chrom_names), repeat(uniform_coverage), repeat(window_size), repeat(interval), repeat(Aa), repeat(Bb), repeat(coverage), repeat(read_len), repeat(seq_error), repeat(out_path)))

def gen_readcounts(tree, chrom_names, bins, num_regions, region_length, uniform_coverage, x0, y0, interval, window_size, coverage, read_len, out_path):
    regions_per_window = round(window_size / region_length)
    [Aa, Bb] = get_alpha_beta(x0, y0)
    leaves = list(tree.iter_leaves())
    final_readcounts = {}

    for cell in leaves:
        prefix = os.path.join(out_path, cell.name)
        with open(prefix + '.pkl', 'rb') as f:
            genome = pickle.load(f)

        full_readcounts = {chrom: [[0 for i in range(num_regions[chrom])], [0 for i in range(num_regions[chrom])]] for chrom in chrom_names}
        for chrom in chrom_names:
            for allele in [0, 1]:
                if len(genome[chrom][allele]) > 0:
                    for homolog in genome[chrom][allele]:
                        while 'X' in homolog:
                            homolog.remove('X')
                        
                        num_windows = int(np.ceil(len(homolog) / regions_per_window))
                        if uniform_coverage:
                            cur_readcounts = [round(poisson.rvs((window_size*coverage) / (2*read_len))) for i in range(num_windows)]
                        else:
                            cur_readcounts = [round(x) for x  in draw_readcounts(num_windows, window_size, interval, Aa, Bb, coverage, read_len)]

                        windows = [list(w) for w in iter_by_chunk(homolog, regions_per_window)]
                        while None in windows[-1]:
                            windows[-1].remove(None)

                        for i,window in enumerate(windows):
                            rc = cur_readcounts[i]
                            cur_window_len = len(window)
                            if cur_window_len != regions_per_window:
                                rc = round(rc * (cur_window_len/regions_per_window))
                            
                            counts = dict(Counter(window))
                            for r,c in counts.items():
                                r_ratio = c / cur_window_len
                                full_readcounts[chrom][allele][r] += (r_ratio * rc)
        

        binned_readcounts = {chrom: [[], []] for chrom in chrom_names}
        for chrom in chrom_names:
            for allele in [0, 1]:
                for i in range(len(bins[chrom])-1):
                    regions = list(range(bins[chrom][i], bins[chrom][i+1]))
                    binned_readcounts[chrom][allele].append(round(sum([full_readcounts[chrom][allele][j] for j in regions])))
        final_readcounts[cell] = binned_readcounts
        os.remove(prefix + '.pkl')

    with open(os.path.join(out_path, 'readcounts.tsv'), 'w+') as f:
        headers = ['CELL', 'chrom', 'start', 'end', 'Acount', 'Bcount']
        f.write('\t'.join(headers) + '\n')

        # Writes to file in order of chroms, bins, cell
        for chrom in chrom_names:
            num_bins = len(bins[chrom]) - 1
            for i in range(num_bins):
                bin_start, bin_end = str(bins[chrom][i]*region_length), str(bins[chrom][i+1]*region_length)
                for cell in leaves:
                    line = [cell.name, chrom, bin_start, bin_end, str(final_readcounts[cell][chrom][0][i]), str(final_readcounts[cell][chrom][1][i])]
                    f.write('\t'.join(line) + '\n')




    


import numpy as np
import scipy.stats
import subprocess
import msprime as ms
import os
from collections import deque, Counter
import copy

class Node:
    def __init__(self, name='', edge_len = 0, parent = None, cell_type='aneuploid'):
        self.name = name
        self.length = edge_len
        self.parent = parent
        self.children = []
        self.events = []
        self.cell_type = cell_type
        #self.whole_chrom_events = []
        self.genome = None
        #self.alterations = None
        self.profile = None

    def __str__(self):
        if self.name:
            return str(self.name)
        else:
            return ''

    def __len__(self):
        return len(self.get_leaves())
    
    #format codes --> 0: leaf names only, 1: leaf names + lengths only, 2: leaf and internal names, 3: leaf and internal names + lengths
    def write_newick(self, terminate=True, format=0):
        if self.is_leaf():
            if format == 0 or format == 2:
                return self.name
            elif format == 1 or format == 3:
                return self.name + ':' + str(self.length)
        else:
            newick_str = '('
            for child in self.children:
                newick_str += child.write_newick(terminate=False, format=format) + ','
            newick_str = newick_str[:-1]
            newick_str += ')'
            if format == 2 or format == 3 or terminate:
                newick_str += self.name
            if (format == 1 or format == 3) and not terminate :
                newick_str += ':' + str(self.length)
        if terminate:
            return newick_str + ';'
        else:
            return newick_str

    def set_name(self, name):
        self.name = name

    def set_len(self, edge_len):
        self.length = float(edge_len)

    def set_parent(self, parent):
        self.parent = parent

    def set_child(self, child):
        if child.parent:
            pass
        self.children.append(child)
        child.parent = self

    def set_sibling(self, sibling):
        self.sibling.append(sibling)

    def set_type(self, cell_type):
        self.cell_type = cell_type

    def inheret(self):
        #self.genome = copy.deepcopy(self.parent.genome)
        #self.alterations = copy.deepcopy(self.parent.alterations)
        self.genome = {}
        chrom_names = list(self.parent.genome.keys())
        for chrom in chrom_names:
            self.genome[chrom] = [[], []]
            for allele in [0, 1]:
                for homolog in self.parent.genome[chrom][allele]:
                    self.genome[chrom][allele].append(homolog.copy())

    def is_leaf(self):
        return len(self.children) == 0
    
    def is_root(self):
        return self.parent is None

    def get_root(self):
        root = self
        while root.parent is not None:
            root = root.parent
        return root

    def add_child(self, name='', edge_len = 0):
        child = Node(name = name, edge_len = float(edge_len), parent = self)
        self.children.append(child)
        return child

    def detach(self):
        if not self.is_root():
            self.parent.children.remove(self)
            self.parent = None

    def iter_postorder(self):
        visit_queue = deque()
        return_queue = deque()
        visit_queue.append(self)

        while visit_queue:
            node = visit_queue.pop()
            return_queue.append(node)
            if not node.is_leaf():
                visit_queue.extend(node.children)
        
        while return_queue:
            node = return_queue.pop()
            yield node

    #level order traversal, i.e. breadth first search from the root. Does include the root itself.
    def iter_descendants(self):
        nodes = deque()
        nodes.append(self)

        while nodes:
            node = nodes.popleft()
            nodes.extend(node.children)
            #if node != self:
            yield node
    
    def iter_preorder(self):
        visit_queue = deque()
        visit_queue.append(self)

        while visit_queue:
            node = visit_queue.pop()
            visit_queue.extend(node.children)
            yield node

    def get_leaves(self):
        return [n for n in self.iter_postorder() if n.is_leaf()]

    def get_height(self):
        if self.is_leaf():
            return 0
        else:
            return max([1 + child.get_height() for child in self.children])

    def get_total_branchlen(self):
        return sum([node.length for node in self.iter_descendants()])

class Tree:
    def __init__(self, root=None, newick=None):
        self.root = root
        self.founder = None
        if newick:
            self.root = str_to_newick(newick)

    def __len__(self):
        return len([leaf for leaf in self.iter_leaves()])

    def print_newick(self, format=0):
        newick_str = self.root.write_newick(format=format)
        print(newick_str)

    def set_founder(self, node):
        self.founder = node
        node.set_type('founder')

    def iter_postorder(self):
        return self.root.iter_postorder()
    
    def iter_descendants(self):
        return self.root.iter_descendants()
    
    def iter_preorder(self):
        return self.root.iter_preorder()
    
    def iter_leaves(self):
        for node in self.iter_descendants():
            if node.is_leaf():
                yield node
    
    def iter_path_to_leaf(self, leaf):
        path = []
        cur_node = leaf
        while not cur_node.is_root():
            path.append(cur_node)
            cur_node = cur_node.parent
        path.append(self.root)
        path.reverse()
        for node in path:
            yield node

    def has_leaf_names(self):
        for leaf in self.iter_leaves():
            if leaf.name == '':
                return False
        return True

    def set_leaf_names(self):
        count = 1
        for leaf in self.iter_leaves():
            if leaf.name == '':
                leaf.name = 'cell' + str(count)
                count += 1

    def set_node_names(self):
        internalcount = 1
        leafcount = 1
        for node in self.iter_descendants():
            if node.is_root():
                if node.cell_type == 'founder':
                    node.name = 'founder'
                else:
                    node.name = 'root'
            elif node.is_leaf():
                #if node.cell_type == 'normal':
                #    node.name = 'normal' + str(leafcount)
                #    leafcount += 1
                #elif node.cell_type == 'pseudonormal':
                #    node.name = 'psd_normal' + str(leafcount)
                #    leafcount += 1
                #else:
                node.name = 'cell' + str(leafcount)
                leafcount += 1
            else:
                if node.cell_type == 'founder':
                    node.name = 'founder'
                else:
                    node.name = 'ancestor' + str(internalcount)
                    internalcount += 1

    def get_tree_height(self):
        return self.root.get_height()

    def get_total_branchlen(self):
        return self.root.get_total_branchlen()

    def ref_node(self, node_name):
        for node in self.iter_descendants():
            if node.name == node_name:
                return node

    def get_mutations(self):
        mutations = []
        for node in self.iter_descendants():
            #mutations += node.whole_chrom_events
            mutations += node.events
        return mutations

    def save(self, file_path, format=0, from_founder=False):
        if from_founder:
            newick_str = self.founder.write_newick(format=format)
        else:
            newick_str = self.root.write_newick(format=format)
        f = open(file_path, 'w+')
        f.write(newick_str)
        f.close()

def str_to_newick(newick_str):
    split_str = newick_str[:-1].split(',')

    cur_node = None
    nodes = []

    for chunk in split_str:
        while chunk[0] == '(':
            new_node = Node()
            if cur_node:
                cur_node.set_child(new_node)
            cur_node = new_node
            chunk = chunk[1:]
        rest = chunk.split(')')
        if ':' in rest[0]:
            idx = rest[0].index(':')
            cur_node.add_child(name=rest[0][:idx], edge_len=rest[0][idx+1:])
        else:
            cur_node.add_child(name=rest[0])
        if len(rest) > 1:
            for part in rest[1:]:
                if ':' in part:
                    idx = part.index(':')
                    cur_node.set_name(part[:idx])
                    cur_node.set_len(part[idx+1:])
                else:
                    cur_node.set_name(part)
                    cur_node.set_len(0)
                if not cur_node.is_root():
                    cur_node = cur_node.parent
    return cur_node

#Tree with random topology by attaching two random leaves. Branch lengths are left undefined.
def gen_tree_random_topology(m, names = None):
    root = Node()
    leaves = [root]

    while len(leaves) < m:
        idx = np.random.randint(0, len(leaves))
        node = leaves.pop(idx)
        c1 = node.add_child()
        c2 = node.add_child()
        leaves.append(c1)
        leaves.append(c2)
    
    count = 0
    for n in root.get_leaves():
        if names:
            n.name = names[count]
        else:
            n.name = 'leaf' + str(count)
        count += 1

    t = Tree(root=root)
    return t

# Creates branch lengths so that each leaf is equally distant from the root.
def add_branchlen_ultrametric(tree):
    node_max_depths = {}
    for node in tree.iter_postorder():
        if node.is_leaf():
            node_max_depths[node] = 1
        else:
            node_max_depths[node] = max([node_max_depths[c] for c in node.children]) + 1
    
    tree_length = float(tree.get_tree_height())
    node_dists = {tree.root: 0.0}
    for node in tree.iter_descendants():
        node.length = (tree_length - node_dists[node.parent]) / node_max_depths[node]
        node_dists[node] = node.length + node_dists[node.parent]
    
# Given a tree topology with existing branch lengths, computes branch lengths uniformly with a random deviation proportional to the tree height.
def add_branchlen_deviation(tree, shape=1):
    for node in tree.iter_descendants():
        x = np.random.gamma(shape)
        node.length = node.length * x

def set_root_branchlen(tree, scale):
    tot = tree.get_total_branchlen()
    tree.root.set_len(tot*scale)

def scale_edge_lengths(tree, place_param):
    leaf_edge_lens = [leaf.length for leaf in tree.founder.get_leaves()]
    avg_leaf_len = sum(leaf_edge_lens) / len(leaf_edge_lens)
    scalar = place_param / avg_leaf_len

    for node in tree.founder.iter_descendants():
        if node == tree.founder:
            node.length = place_param
        else:
            node.length = node.length * scalar

# OUTDATED
# Calls the ms binary from the given path. 
def call_ms(num_cells, out_path, growth_rate):
    seed_vals = np.random.randint(1, 100000, 3)
    temp_path = os.path.join(out_path, 'temp_log')
    f = open(temp_path, 'w+')
    if growth_rate == 0:
        call = subprocess.call(['mspms', str(num_cells), '1', '-T', '-seeds', str(seed_vals[0]), str(seed_vals[1]), str(seed_vals[2])], stdout=f)
    else:
        call = subprocess.call(['mspms', str(num_cells), '1', '-G', str(growth_rate), '-T', '-seeds', str(seed_vals[0]), str(seed_vals[1]), str(seed_vals[2])], stdout=f)
    f.close()
    with open(temp_path) as f:
        line = f.readline()
        while line[0] != '(':
            line = f.readline()
        tree_str = line[:-1]

    os.remove(temp_path)
    return tree_str

def gen_tree_standard(num_cells, growth_rate):
    demo = ms.Demography()
    demo.add_population(
        initial_size=0.5,
        growth_rate=growth_rate
    )
    ts = ms.sim_ancestry(
        samples=num_cells,
        demography=demo,
        model=ms.StandardCoalescent(),
        ploidy=1
    )
    tree_str = ts.first().as_newick()
    return tree_str

def gen_tree_sweep(num_cells, growth_rate, num_sweep, s):
    demo = ms.Demography()
    demo.add_population(
        initial_size=0.5,
        growth_rate=growth_rate
    )
    sampleset = ms.SampleSet(num_cells, ploidy=1)
    L = 1
    Ne = min(1e8, num_cells*1e4)

    models = []
    for i in range(num_sweep):
        models.append(ms.StandardCoalescent(duration=np.random.uniform()))
        models.append(
            ms.SweepGenicSelection(
                #position=np.random.randint(1, L-1),
                position=0,
                start_frequency=1 / (2*Ne),
                end_frequency=1 - 1 / (2*Ne),
                s = s,
                dt = 1/Ne
            )
        )
    models.append(ms.StandardCoalescent())

    ts = ms.sim_ancestry(
        samples=[sampleset],
        demography=demo,
        model=models,
        sequence_length=L
    )
    tree_str = ts.first().as_newick()
    return tree_str


def make_tumor_tree(tree_type, num_cells, normal_frac, pseudonormal_frac, root_events, out_path, growth_rate, tree_path, num_sweep, s):
    num_normal = round(num_cells * normal_frac)
    if root_events == 0:
        num_pseudonormal = 0
    else:
        num_pseudonormal = round(num_cells * pseudonormal_frac)
    num_aneuploid = num_cells - num_normal - num_pseudonormal

    if num_aneuploid > 0:
        if tree_type == 0:
            if num_sweep > 0:
                tree_str = gen_tree_sweep(num_aneuploid, growth_rate, num_sweep, s)
            else:
                tree_str = gen_tree_standard(num_aneuploid, growth_rate)
            #tree_str = call_ms(num_aneuploid, out_path, growth_rate)
            tree = Tree(newick=tree_str)
        elif tree_type == 1:
            tree = gen_tree_random_topology(num_aneuploid)
        elif tree_type == 2:
            f = open(tree_path)
            tree_str = f.readline()
            f.close()
            tree = Tree(newick=tree_str)
    else:
        tree = Tree()
    
    # Create a new root node and add normal cells as children.
    if num_normal > 0:
        norm_root = Node()
        norm_root.cell_type = 'normal'
        for i in range(num_normal):
            n = Node()
            n.cell_type = 'normal'
            norm_root.set_child(n)
    
    # Handling diverse pseudonormal cells
    if num_pseudonormal > 0:
        # Create a section of the tree for pseudonormals. Distribute them to have anywhere from 1 to 'root_events' number of events. Form the tree accordingly.
        psnorm_root = Node()
        psnorm_root.cell_type = 'pseudonormal'
        placements = Counter(list(np.random.randint(1, root_events+1, size=num_pseudonormal)))
        ps_keys = sorted(list(placements.keys()))

        psnorm_root.events = [None for i in range(ps_keys[0])]
        for i in range(placements[ps_keys[0]]):
            n = Node()
            n.set_type('pseudonormal')
            psnorm_root.set_child(n)

        cur_node = psnorm_root
        cur_mut = ps_keys[0]
        for k in ps_keys[1:]:
            next_int = Node()
            next_int.cell_type = 'pseudonormal'
            next_int.events = [None for i in range(k - cur_mut)]
            cur_mut = k
            for i in range(placements[k]):
                n = Node()
                n.cell_type = 'pseudonormal'
                next_int.set_child(n)
            cur_node.set_child(next_int)
            cur_node = next_int
        if num_aneuploid > 0:
            tree.root.events = [None for i in range(root_events - cur_mut)]

    # Make updates to the tree structure accordingly
    if num_aneuploid > 0:
        tree.set_founder(tree.root)
        if num_normal > 0 and num_pseudonormal == 0:
            norm_root.set_child(tree.root)
            tree.root = norm_root
        elif num_normal == 0 and num_pseudonormal > 0:
            cur_node.set_child(tree.root)
            tree.root = psnorm_root
        elif num_normal > 0 and num_pseudonormal > 0:
            norm_root.set_child(psnorm_root)
            cur_node.set_child(tree.root)
            tree.root = norm_root
        else:
            if root_events == 0:
                tree.root.events = []
            else:
                tree.root.events = [None for i in range(root_events)]
    else:
        if num_normal > 0:
            tree.root = norm_root
            if num_pseudonormal > 0:
                tree.root.set_child(psnorm_root)
        elif num_pseudonormal > 0:
            tree.root = psnorm_root

    return tree

def select_clones(tree, num_clones, criteria, mu, sig):
    ancestral_aneuploids = {}
    for node in tree.founder.iter_descendants():
        if not node == tree.founder and not node.is_leaf():
            ancestral_aneuploids[node] = len(node)

    nodes, sizes = zip(*ancestral_aneuploids.items())

    if criteria == 0:
        if not mu:
            sizes = [s/sum(sizes) for s in sizes]
        else:
            if not sig:
                sig = mu*0.1
            x = scipy.stats.norm(mu, sig)
            sizes = [x.pdf(s) for s in sizes]
            sizes = [s/sum(sizes) for s in sizes]
    elif criteria == 1:
        sizes = [x.length for x in nodes]
        sizes = [x/sum(sizes) for x in sizes]
    
    clone_founders = np.random.choice(nodes, size=num_clones, p=sizes, replace=False)

    for clone in clone_founders:
        clone.cell_type = 'clone'

    return list(clone_founders)



import numpy as np

def choose_noise_model(choice, tree, chrom_names, err_rate, dub_rate):
    if choice == 1:
        add_noise_simple(tree, chrom_names, err_rate)
    elif choice == 2:
        add_noise_normal(tree, chrom_names, err_rate)
    elif choice == 3:
        add_noise_boundary(tree, chrom_names, err_rate)
    elif choice == 4:
        add_noise_sequence(tree, chrom_names, err_rate, dub_rate)
    elif choice == 5:
        add_noise_boundary(tree, chrom_names, err_rate)
        add_noise_normal(tree, chrom_names, dub_rate)

def add_noise_simple(tree, chrom_names, err_rate):
    for leaf in tree.iter_leaves():
        for chrom in chrom_names:
            for allele in [0, 1]:
                for b in range(len(leaf.profile[chrom][allele])):
                    if np.random.uniform() < err_rate:
                        if leaf.profile[chrom][allele][b] != 0:
                            if np.random.binomial(1, 0.5) == 0:
                                leaf.profile[chrom][allele][b] += 1
                            else:
                                leaf.profile[chrom][allele][b] -= 1

def add_noise_normal(leaf, chrom_names, err_rate):
    for chrom in chrom_names:
        for allele in [0, 1]:
            for b in range(len(leaf.profile[chrom][allele])):
                leaf.profile[chrom][allele][b] = round(np.random.normal(leaf.profile[chrom][allele][b], err_rate*leaf.profile[chrom][allele][b]))

def add_noise_boundary(leaf, chrom_names, err_rate):
    for chrom in chrom_names:
        for allele in [0, 1]:
            profile = leaf.profile[chrom][allele]
            i, seg_length = 1, 1
            prev_num = profile[0]
            while i < len(profile):
                if profile[i] != prev_num:
                    new_length = max(round(np.random.normal(seg_length, seg_length*err_rate)), 1)
                    if new_length <= seg_length:
                        for j in range(i - seg_length + new_length, i):
                            profile[j] = profile[i]
                        seg_length = 1
                        prev_num = profile[i]
                        i += 1
                    else:
                        for j in range(i, min(i + new_length - seg_length, len(profile))):
                            profile[j] = prev_num
                        seg_length = 1
                        i += new_length - seg_length
                else:
                    seg_length += 1
                    i += 1

def add_noise_sequence(tree, chrom_names, cov_rate, dub_rate):
    #select doublets
    doublets = {}
    cell_names = list(tree.iter_leaves())
    for i in range(len(cell_names)):
        if np.random.uniform() < dub_rate:
            pooled_cell = cell_names[np.random.choice([j for j in range(len(cell_names)) if j != i])]
            doublets[cell_names[i]] = pooled_cell
    for chrom in chrom_names:
        for allele in [0, 1]:
            for doublet, pooled_cell in doublets.items():
                profile = doublet.profile[chrom][allele]
                alt_profile = pooled_cell.profile[chrom][allele]
                for i in range(len(profile)):
                    if np.random.binomial(1, 0.5) == 0:
                        profile[i] = alt_profile[i]
                    
    #coverage non-uniformity
    for leaf in tree.iter_leaves():
        for chrom in chrom_names:
            for allele in [0, 1]:
                profile = leaf.profile[chrom][allele]
                cov_chart = gen_coverage(len(profile), np.random.randint(4, 7))
                for i in range(len(profile)):
                    copy_num = profile[i]
                    coverage = 2*cov_chart[i]
                    if coverage < 1:
                        scale = 1 - (coverage*cov_rate)
                        profile[i] = round(copy_num*scale)
                    elif coverage > 1:
                        scale = 1 + (coverage*cov_rate)
                        profile[i] = round(copy_num*scale)

# Jitter and boundary model combined
def add_noise_mixed(tree, chrom_names, boundary_err_rate, jitter_err_rate):
    for leaf in tree.iter_leaves():
        if leaf.cell_type != 'normal':
            add_noise_boundary(leaf, chrom_names, boundary_err_rate)
            add_noise_normal(leaf, chrom_names, jitter_err_rate)


def bezier_coef(points):
    n = len(points) - 1

    C = 4 * np.identity(n)
    np.fill_diagonal(C[1:], 1)
    np.fill_diagonal(C[:, 1:], 1)
    C[0, 0] = 2
    C[n - 1, n - 1] = 7
    C[n - 1, n - 2] = 2

    P = [2 * (2 * points[i] + points[i + 1]) for i in range(n)]
    P[0] = points[0] + 2 * points[1]
    P[n - 1] = 8 * points[n - 1] + points[n]

    A = np.linalg.solve(C, P)
    B = [0] * n
    for i in range(n - 1):
        B[i] = 2 * points[i + 1] - A[i + 1]
    B[n - 1] = (A[n - 1] + points[n]) / 2

    return A, B

def single_cubic_bezier(a, b, c, d):
    return lambda t: np.power(1 - t, 3) * a + 3 * np.power(1 - t, 2) * t * b + 3 * (1 - t) * np.power(t, 2) * c + np.power(t, 3) * d

def multiple_cubic_bezier(points):
    A, B = bezier_coef(points)
    return [single_cubic_bezier(points[i], A[i], B[i], points[i + 1]) for i in range(len(points) - 1)]

def gen_random_interval(num_bins, scale):
    x = [i for i in range(0, num_bins, round(num_bins/scale))]
    if x[-1] != num_bins-1:
        if num_bins - x[-1] >= round(num_bins/scale)/2:
            x.append(num_bins-1)
        else:
            x[-1] = num_bins-1
    y = [np.random.uniform(0, 1) for i in range(len(x))]
    points = [np.array(p) for p in zip(x, y)]
    return points
    
def gen_coverage(num_bins, scale):
    points = gen_random_interval(num_bins, scale)
    A, B = bezier_coef(points)
    curves = [single_cubic_bezier(points[i], A[i], B[i], points[i + 1]) for i in range(len(points) - 1)]

    new_points = []
    for i in range(len(points) - 1):
        f = curves[i]
        gaps = points[i+1][0] - points[i][0] + 1
        coords = [f(t) for t in np.linspace(0, 1, int(gaps))]
        if i == 0:
            new_points.append((round(coords[0][0]), coords[0][1]))
        new_points += [(round(i[0]), i[1]) for i in coords[1:]]
    
    new_points.sort(key=lambda x: x[0])
    new_points = [max(min(i[1], 1), 0) for i in new_points]
    return new_points



import os
import sys

class InputError(Exception):
    def __init__(self, text):
        self.message = str(text)
        super().__init__(self.message)

class ChromNameError(Exception):
    def __init__(self):
        self.message = "Chromosome names in reference do not match the alternate and/or precomputed hg38 names. Check chromosome names or consider disabling the --use-hg38-static parameter with this reference."
        super().__init__(self.message)

class ChromNumError(Exception):
    def __init__(self):
        self.message = "Cannot have more than 22 chroms with the --use-hg38-static parameter toggled."
        super().__init__(self.message)

class ChromSizeError(Exception):
    def __init__(self):
        self.message = "Detected chromosomes in reference smaller than window size. Ensure the reference contains the desired chromosome sequences and remove alternate sequences if necessary."
        super().__init__(self.message)

class ModeError(Exception):
    def __init__(self):
        self.message = "The simulation mode needs to be specified correctly. Select 0 for CNP, 1 for seq, or 2 for both."
        super().__init__(self.message)

def convert(val):
    if val == 'True':
        return True
    if val == 'False':
        return False
    constructors = [int, float]
    for c in constructors:
        try:
            return c(val)
        except ValueError:
            pass
    return val

def handle_args(arguments):
    if arguments.param_file:
        args = {}
        f = open(arguments.param_file)
        lines = f.readlines()
        f.close()

        for line in lines:
            if line[0] == '[':
                param, val = line[line.index('(')+1:line.index(')')], line[line.index('[')+1:line.index(']')]
                args[param] = convert(val)

    else:
        args = vars(arguments)

    if not args['normal_fraction'] + args['pseudonormal_fraction'] <= 1.0:
        raise InputError('Normal and pseudornormal fractions cannot sum to more than 1.')
    return args

#gets chromosome and arm lengths from cytoband file
def hg38_chrom_lengths_from_cytoband_file(file_path, include_allosomes=False, include_arms=False):
    chrom_lens, arm_ratios = {}, {}

    f = open(file_path)
    lines = f.readlines()
    f.close

    for line in lines:
        info = line.strip().split('\t')
        chrom, start, end, arm = info[0], int(info[1]), int(info[2]), info[3][0]
        if chrom not in chrom_lens:
            chrom_lens[chrom] = {'p': 0, 'q': 0}
        chrom_lens[chrom][arm] = end

    if not include_allosomes:
        del chrom_lens['chrX']
        del chrom_lens['chrY']

    for chrom in chrom_lens:
        arm_ratios[chrom] = chrom_lens[chrom]['p'] / chrom_lens[chrom]['q']
        chrom_lens[chrom] = chrom_lens[chrom]['q']

    if include_arms:
        return chrom_lens, arm_ratios
    else:
        return chrom_lens
    
def hg38_chrom_lengths_from_cytoband(include_allosomes=False, include_arms=False):
    chrom_lens = {'chr1': 249250621, 'chr10': 135534747, 'chr11': 135006516, 'chr12': 133851895, 'chr13': 115169878, 'chr14': 107349540, 'chr15': 102531392, 'chr16': 90354753, 'chr17': 81195210, 'chr18': 78077248, 'chr19': 59128983, 'chr2': 243199373, 'chr20': 63025520, 'chr21': 48129895, 'chr22': 51304566, 'chr3': 198022430, 'chr4': 191154276, 'chr5': 180915260, 'chr6': 171115067, 'chr7': 159138663, 'chr8': 146364022, 'chr9': 141213431, 'chrX': 155270560, 'chrY': 59373566}
    arm_ratios = {'chr1': 0.5015, 'chr10': 0.2966, 'chr11': 0.39776, 'chr12': 0.26746, 'chr13': 0.15542, 'chr14': 0.16395, 'chr15': 0.18531, 'chr16': 0.40507, 'chr17': 0.29558, 'chr18': 0.22029, 'chr19': 0.44817, 'chr2': 0.38364, 'chr20': 0.43633, 'chr21': 0.27426, 'chr22': 0.28652, 'chr3': 0.45954, 'chr4': 0.26366, 'chr5': 0.26753, 'chr6': 0.35649, 'chr7': 0.3764, 'chr8': 0.31155, 'chr9': 0.34699, 'chrX': 0.39029, 'chrY': 0.21053}
    if not include_allosomes:
        del chrom_lens['chrX']
        del chrom_lens['chrY']
        del arm_ratios['chrX']
        del arm_ratios['chrY']
    if include_arms:
        return chrom_lens, arm_ratios
    else:
        return chrom_lens
    
def check_chrom_lengths(chrom_lens, window_size):
    for c,l in chrom_lens.items():
        if l < window_size:
            return False
    return True

# Summary sim stats 
def summary(tree, out_path, num_chroms, WGD, region_length):
    pass

def record_cell_types(tree, out_path, chrom_level_event, super_clone_rate, clone_founders=[]):
    f = open(out_path, 'w+')
    if super_clone_rate and chrom_level_event:
        #for c in tree.founder.children:
        #    if c not in clone_founders:
        #        clone_founders.append(c)
        clone_founders.extend(tree.founder.children)
    if len(clone_founders) > 0:
        clone_founders.sort(key=lambda x: len(x), reverse=True)
        cloneid = 0
        membership = {}
        for n in tree.iter_descendants():
            membership[n] = n.cell_type
        for clone in clone_founders:
            cloneid += 1
            for n in clone.iter_descendants():
                if n == clone:
                    membership[n] = 'clone' + str(cloneid) + '_founder'
                else:
                    membership[n] = 'clone' + str(cloneid)
        for n in tree.iter_descendants():
            f.write(n.name + '\t' + membership[n] + '\n')
    else:
        for n in tree.iter_descendants():
            f.write(n.name + '\t' + n.cell_type + '\n')
    f.close()
    return clone_founders

def record_clone_events(tree, out_path, super_clone_rate, clone_founders):
    f = open(out_path, 'w+')
    cloneids = {clone_founders[i]: 'clone' + str(i+1) for i in range(len(clone_founders))}
    eventdict = {0: 'del', 1: 'dup'}
    scaledict = {1: {None: 'whole'}, 2: {'p': 'p', 'q': 'q', None: 'Other'}}
    f.write('\t'.join(['cellname', 'cloneid', 'chrom', 'allele', 'scale', 'event']) + '\n')
    for e in tree.founder.events:
        if e.category != 0:
            f.write('\t'.join([str(x) for x in ['founder', 'None', e.chrom, e.allele, scaledict[e.category][e.arm], eventdict[e.event]]]) + '\n')
    for c in clone_founders:
        for e in c.events:
            if e.category != 0:
                f.write('\t'.join([str(x) for x in [c.name, cloneids[c], e.chrom, e.allele, scaledict[e.category][e.arm], eventdict[e.event]]]) + '\n')
    f.close()

def record_events(tree, region_length, out_path):
    f = open(out_path, 'w+')
    eventdict = {0: 'del', 1: 'gain'}
    f.write('\t'.join(['cellname', 'chrom', 'allele', 'tumor_start_idx', 'ref_start_idx', 'length', 'event', 'copies']) + '\n')
    for n in tree.iter_descendants():
        for e in n.events:
            if e.category == 0:
                f.write('\t'.join([str(x) for x in [n.name, e.chrom, e.allele, e.start*region_length + 1, e.ref_idx*region_length + 1, e.length*region_length, eventdict[e.event], e.copies]]) + '\n')
    f.close()

def get_size(obj, seen=None):
    """Recursively finds size of objects"""
    size = sys.getsizeof(obj)
    if seen is None:
        seen = set()
    obj_id = id(obj)
    if obj_id in seen:
        return 0
    # Important mark as seen *before* entering recursion to gracefully handle
    # self-referential objects
    seen.add(obj_id)
    if isinstance(obj, dict):
        size += sum([get_size(v, seen) for v in obj.values()])
        size += sum([get_size(k, seen) for k in obj.keys()])
    elif hasattr(obj, '__dict__'):
        size += get_size(obj.__dict__, seen)
    elif hasattr(obj, '__iter__') and not isinstance(obj, (str, bytes, bytearray)):
        size += sum([get_size(i, seen) for i in obj])
    return size


from pyfaidx import Fasta
import numpy as np
import subprocess
import os
import pickle

def read_fasta(input_fasta, chrom_lens_only = False):
    ref = Fasta(input_fasta, one_based_attributes=False)
    chrom_lens = {}
    for chrom in ref.keys():
        chrom_lens[chrom] = len(ref[chrom])
    if chrom_lens_only:
        return chrom_lens
    else:
        return ref, chrom_lens

def make_alt_ref(ref):
    pass

def build_cell_ref(genome, ref, chrom_names, regions_per_chrom, region_length, allele, prefix):
    if isinstance(genome, str):
        with open(genome, 'rb') as f:
            genome = pickle.load(f)
    ref_name = prefix + '_allele' + str(allele) + '.fa'
    chrom_lens = {}
    f = open(ref_name, 'w+')
    for chrom in chrom_names:
        chrom_lens[chrom] = 0
        num_regions = regions_per_chrom[chrom]
        if len(genome[chrom][allele]) > 0:
            f.write('>' + chrom + '\n')
            for homolog in genome[chrom][allele]:
                while 'X' in homolog:
                    homolog.remove('X')
                for r in homolog:
                    if r == num_regions - 1:
                        f.write(ref[chrom][r*region_length:].seq)
                        chrom_lens[chrom] += len(ref[chrom][r*region_length:].seq)
                    else:
                        f.write(ref[chrom][r*region_length:(r+1)*region_length].seq)
                        chrom_lens[chrom] += region_length
            f.write('\n')
    f.close()
    call = subprocess.run(['samtools', 'faidx', ref_name])
    #cur_ref = Fasta(ref_name)
    return ref_name, chrom_lens



from setuptools import setup, find_packages

setup(
    name='CNAsim',
    version='1.3.5',
    author='Samson Weiner',
    author_email='samson.weiner@uconn.edu',
    description='CNAsim is a software package for simulation of single-cell CNA data from tumors.',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/samsonweiner/CNAsim',
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent'
    ],
    packages=find_packages(),
    entry_points={
        'console_scripts': [
            'cnasim=CNAsim.main:main',
        ],
    },
    python_requires='>=3.7'
)


#!/usr/bin/env python3

# Copyright (C) 2023 Samson Weiner (samson.weiner@uconn.edu) and
# Mukul S. Bansal (mukul.bansal@uconn.edu).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see http://www.gnu.org/licenses/.


import argparse
import os
import time

from .tree import *
from .sim_genomes import init_diploid_genome, evolve_tree, prepare_ancestral_profiles
from .sequence import read_fasta
from .reads import gen_reads, gen_readcounts
from .noise import add_noise_mixed
from .format_profiles import save_CN_profiles_leaves, save_CN_profiles_ancestors
from .utilities import *


def parse_args():
    #Arguments
    parser = argparse.ArgumentParser()
    parser.add_argument('-m', '--mode', type=int, default=None, help='Main simulator mode for generating data. 0: CNPs only, 1: readcounts & CNPs, 2: sequencing reads & CNPs')
    parser.add_argument('-o', '--out-path', type=str, default='CNAsim_output/', help='Path to output directory.')
    parser.add_argument('-t', '--tree-type', type=int, default=0, help='0: coalescence, 1: random, 2: from file (use -T to specify file path).')
    parser.add_argument('-T', '--tree-path', type=str, default=None, help='Path to input tree.')
    parser.add_argument('-g', '--growth-rate', type=float, default=0.003785, help='Exponential growth rate for standard coalescent.')
    parser.add_argument('-s', '--num-sweep', type=int, default=0, help='Number of selective sweeps in coalescent.')
    parser.add_argument('-s1', '--selection-strength', type=float, default=0.01, help='Parameter controlling the strength of selection during the sweeps.')
    parser.add_argument('-n', '--num-cells', type=int, default=250, help='Number of observed cells in sample.')
    parser.add_argument('-n1', '--normal-fraction', type=float, default=0, help='Proportion of cells that are normal.')
    parser.add_argument('-n2', '--pseudonormal-fraction', type=float, default=0, help='Proportion of cells that are pseudonormal cells.')
    parser.add_argument('-c', '--num-clones', type=int, default=0, help='Number of ancestral nodes to select as clonal founders.')
    parser.add_argument('-c1', '--clone-criteria', type=int, default=0, help='Criteria to choose clonal ancesters. 0: proportional to number of leaves in subtree. 1: proportional to edge length')
    parser.add_argument('-c2', '--clone-mu', type=float, default=None, help='Mean number of leaves in a subclone. Must select 0 for clone-criteria.')
    parser.add_argument('-c3', '--clone-sd', type=float, default=None, help='SD in the number of leaves in a subclone. Must select 0 for clone-criteria.')
    parser.add_argument('-p', '--placement-type', type=int, default=0, help='Number of CNAs per edge. 0: draw from a Poisson with fixed mean, 1: draw from a Poisson with mean prop to edge length, 2: fixed per edge')
    parser.add_argument('-p1', '--placement-param', type=float, default=2, help='Parameter for placement choice.')
    parser.add_argument('-k', '--region-length', type=int, default=1000, help='Region length in bp. Essentially controls the resolution of the simulated genome.')
    parser.add_argument('-l', '--cn-length-mean', type=int, default=5000000, help='Mean copy number event length in bp.')
    parser.add_argument('-l1', '--min-cn-length', type=int, default=1000, help='Minimum copy number event length in bp. Should be at minimum the region length and less than the mean.')
    parser.add_argument('-a', '--cn-copy-param', type=float, default=0.5, help='Parameter in the geometric to select number of copies.')
    parser.add_argument('-b', '--cn-event-rate', type=float, default=0.5, help='Probability an event is an amplification. Deletion rate is 1 - amp rate.')
    parser.add_argument('-j', '--founder-event-mult', type=int, default=10, help='Multiplier for the number of events along edge into founder cell.')
    parser.add_argument('-w', '--WGD', action='store_true', help='Include WGD.')
    parser.add_argument('-v', '--chrom-level-event', action='store_true', help='Include chromosomal alterations.')
    parser.add_argument('-q', '--chrom-arm-rate', type=float, default=0.75, help='Probability that a chromosomal event is a chromosome-arm event.')
    parser.add_argument('-i', '--chrom-rate-founder', type=float, default=2, help='Parameter in poisson for number of chromosome-level events along the edge into the founder cell.')
    parser.add_argument('-i1', '--chrom-rate-super-clone', type=float, default=1, help='Parameter in poisson for number of chromosome-level events along the edges out of the founder cell.')
    parser.add_argument('-i2', '--chrom-rate-clone', type=float, default=1, help='Parameter in poisson for number of chrom-level events for clonal nodes.')
    parser.add_argument('-u', '--chrom-event-type', type=float, default=0.5, help='Probability that a chromosomal event is a duplication.')
    parser.add_argument('-N', '--num-chromosomes', type=int, default=22, help='Number of chromosomes if run in mode 0.')
    parser.add_argument('-L', '--chrom-length', type=int, default=100000000, help='Length of chromosomes in bp if not using hg38 static.')
    parser.add_argument('-A', '--chrom-arm-ratio', type=float, default=0.5, help='If not using hg38 static, ratio of length within the p-arm.')
    parser.add_argument('-B', '--bin-length', type=int, default=1000000, help='Resolution of copy number profiles in bp.')
    parser.add_argument('-E1', '--error-rate-1', type=float, default=0, help='Error rate for the boundary model.')
    parser.add_argument('-E2', '--error-rate-2', type=float, default=0, help='Error rate for the jitter model.')
    #parser.add_argument('-O', '--output-clean-CNP', action='store_true', help='Output the clean CNPs in addition to the noisy ones.')
    parser.add_argument('-U', '--use-hg38-static', action='store_true', help='Use hg38 chromosome information. Excludes sex chromosomes chrX and chrY.')
    #parser.add_argument('-A', '--include-autosomes', action='store_true', help='Include autosomes in genome. Must either use hg38 static with number of chromosomes > 22, or pass a reference genome with autosomes.')
    parser.add_argument('-r1', '--reference', type=str, default=None, help='Path to input reference genome as the primary haplotype in fasta format. Will be duplicated as both haplotypes if an alternate is not provided.')
    parser.add_argument('-r2', '--alt-reference', type=str, default=None, help='Path to an alternate reference genome to be used as a secondary haplotype, also in fasta format.')
    #parser.add_argument('-D', '--chrom-names-path', type=str, default=None, help='Path to file containing chromosomes names to use in reference file(s). By default, will use all chromosomes present in reference except chrX and chrY.')
    parser.add_argument('-M', '--use-uniform-coverage', action='store_true', help='Use uniform coverage across the genome.')
    parser.add_argument('-X', '--lorenz-x', type=float, default=0.5, help='x-coordinate of point on lorenz curve if using non-uniform coverage.')
    parser.add_argument('-Y', '--lorenz-y', type=float, default=0.4, help='y-coordinate of point on lorenz curve if using non-uniform coverage.')
    parser.add_argument('-W', '--window-size', type=int, default=1000000, help='Number of base pairs to generate reads for in each iteration.')
    parser.add_argument('-I', '--interval', type=int, default=3, help='Initializes a point in the coverage distribution every interval number of windows.')
    parser.add_argument('-C', '--coverage', type=float, default=0.1, help='Average sequencing coverage across the genome.')
    parser.add_argument('-R', '--read-length', type=int, default=150, help='Paired-end short read length.')
    parser.add_argument('-S', '--seq-error', type=float, default=0.02, help='Per base error rate for generating sequence data.')
    parser.add_argument('-P', '--processors', type=int, default=1, help='Number of processes to use for generating reads in parallel.')
    parser.add_argument('-d', '--disable-info', action='store_true', help='Do not output simulation log, cell types, or ground truth events.')
    parser.add_argument('-F', '--param-file', type=str, default=None, help='Path to parameter file to specify parameters instead of the command line. Must conform to the sample format.')
    arguments = parser.parse_args()

    return handle_args(arguments)

def main():
    args = parse_args()

    if args['mode'] not in [0, 1, 2]:
        raise ModeError

    ## Initialize output directory
    print('Output directory:', os.path.abspath(args['out_path']))
    if not os.path.isdir(args['out_path']):
        os.makedirs(args['out_path'])

    ## Log parameters
    if not args['disable_info']:
        start = time.time()
        log = open(os.path.join(args['out_path'], 'log.txt'), 'w+')
        for k,v in args.items():
            log.write(k + ': ' + str(v) + '\n')

    ## Create tree structure
    print('Preparing ground truth tree...')
    founder_events = int(args['placement_param'] * args['founder_event_mult'])
    tree = make_tumor_tree(args['tree_type'], args['num_cells'], args['normal_fraction'], args['pseudonormal_fraction'], founder_events, args['out_path'], args['growth_rate'], args['tree_path'], args['num_sweep'], args['selection_strength'])
    clone_founders = []
    if args['num_clones'] > 0:
        clone_founders = select_clones(tree, args['num_clones'], args['clone_criteria'], args['clone_mu'], args['clone_sd'])
        args['chrom_level_event'] = True
    tree.set_node_names()
    tree.save(os.path.join(args['out_path'], 'tree.nwk'), format=3)

    if not args['disable_info']:
        clone_founders = record_cell_types(tree, os.path.join(args['out_path'], 'cell_types.tsv'), args['chrom_level_event'], args['chrom_rate_super_clone'], clone_founders=clone_founders)

    if args['placement_type'] == 1:
        scale_edge_lengths(tree, args['placement_param'])

    ## Initialize genome
    print('Initializing reference genome...')
    if args['reference']:
        if not os.path.isfile(args['reference']):
            raise InputError("Cannot access input " + str(args['reference']))
        ref1, chrom_lens = read_fasta(args['reference'])
        if args['use_hg38_static']:
            all_chroms = list(chrom_lens.keys())
            chrom_names = [f'chr{i}' for i in range(1, 23)]
            for c in all_chroms:
                if c not in chrom_names:
                    chrom_lens.pop(c, None)
        chrom_names = list(chrom_lens.keys())
        print(f'Number of chromosomes in reference: {len(chrom_names)}')
        if not check_chrom_lengths(chrom_lens, args['window_size']):
            if args['mode'] == 2:
                raise ChromSizeError()

        if args['alt_reference']:
            if not os.path.isfile(args['alt_reference']):
                raise InputError("Cannot access input " + str(args['alt_reference']))
            ref2, alt_chrom_lens = read_fasta(args['alt_reference'])
            if args['use_hg38_static']:
                all_chroms = list(chrom_lens.keys())
                alt_chrom_names = [f'chr{i}' for i in range(1, 23)]
                for c in all_chroms:
                    if c not in alt_chrom_names:
                        alt_chrom_lens.pop(c, None)
            if set(chrom_names) != set(alt_chrom_lens.keys()):
                raise ChromNameError()
        else:
            ref2 = ref1
    else:
        if args['mode'] == 2:
            raise InputError("Cannot access input " + str(args['reference']))
        chrom_names = ['chr' + str(i+1) for i in range(args['num_chromosomes'])]
        chrom_lens = dict(zip(chrom_names, [args['chrom_length'] for i in range(args['num_chromosomes'])]))

    if args['use_hg38_static']:
        file_path = os.path.realpath(__file__)
        sim_dir_path, filename = os.path.split(file_path)
        if not args['reference']:
            chrom_lens, arm_ratios = hg38_chrom_lengths_from_cytoband(include_allosomes=False, include_arms=True)
            if args['num_chromosomes'] > 22:
                raise ChromNumError()
        else:
            temp, arm_ratios = hg38_chrom_lengths_from_cytoband(include_allosomes=False, include_arms=True)
            for chrom in chrom_names:
                if chrom not in arm_ratios:
                    raise ChromNameError()
    else:
        arm_ratios = args['chrom_arm_ratio']
    
    normal_diploid_genome, num_regions = init_diploid_genome(args['region_length'], chrom_names, chrom_lens, arm_ratios)
    tree.root.genome = normal_diploid_genome

    ## Evolve genomes along tree
    print('Generating genomes, events, and profiles...')
    #if args['mode'] == 0 or args['mode'] == 2:
    regions_per_bin = np.floor(args['bin_length']/args['region_length'])
    bins = {}
    for chrom in chrom_names:
        bins[chrom] = [k for k in range(num_regions[chrom]) if k % regions_per_bin == 0]
        if (num_regions[chrom] - 1) - bins[chrom][-1] < regions_per_bin/2:
            bins[chrom][-1] = num_regions[chrom]
        else:
            bins[chrom].append(num_regions[chrom])

    evolve_tree(tree.root, args, chrom_names, num_regions, bins=bins)
    prepare_ancestral_profiles(tree, args, chrom_names, num_regions, bins=bins)

    ## Generate data
    #if args['mode'] == 0 or args['mode'] == 2:
    print('Formating and saving profiles')
    if args['error_rate_1'] != 0 or args['error_rate_2'] != 0:
        save_CN_profiles_leaves(tree, chrom_names, bins, args['region_length'], os.path.join(args['out_path'], 'clean_profiles.tsv'))
        add_noise_mixed(tree, chrom_names, args['error_rate_1'], args['error_rate_2'])
    save_CN_profiles_leaves(tree, chrom_names, bins, args['region_length'], os.path.join(args['out_path'], 'profiles.tsv'))
    save_CN_profiles_ancestors(tree, chrom_names, bins, args['region_length'], os.path.join(args['out_path'], 'ancestral_profiles.tsv'))

    if args['mode'] == 1:
        print('Generating read counts...')
        gen_readcounts(tree, chrom_names, bins, num_regions, args['region_length'], args['use_uniform_coverage'], args['lorenz_x'], args['lorenz_y'], args['interval'], args['window_size'], args['coverage'] / 2, args['read_length'], args['out_path'])
    
    if args['mode'] == 2:
        print('Generating synthetic sequencing reads...')
        gen_reads(ref1, ref2, num_regions, chrom_names, tree, args['use_uniform_coverage'], args['lorenz_x'], args['lorenz_y'], args['region_length'], args['interval'], args['window_size'], args['coverage'] / 2, args['read_length'], args['seq_error'], args['out_path'], args['processors'])

    ## Logging information
    if not args['disable_info']:
        total_time = round(time.time() - start)
        log.write('\nTime: ' + str(total_time) + '\n')
        log.close()

    if not args['disable_info']:
        if (args['num_clones'] > 0 or args['chrom_rate_super_clone']) and args['chrom_level_event']:
            record_clone_events(tree, os.path.join(args['out_path'], 'clone_events.tsv'), args['chrom_rate_super_clone'], clone_founders)
        record_events(tree, args['region_length'], os.path.join(args['out_path'], 'focal_events.tsv'))

if __name__ == '__main__':
    main()


import numpy as np
from .tree import Tree
from collections import Counter
import os
import pickle


# Categories --> 0: focal, 1: whole-chromosomal, 2: chromosome-arm
class CNV():
    def __init__(self, cell = None, category=None, chrom=None, allele=None, homolog=None, arm=None, start=None, length=None, event=None, copies=None, ref_idx=None):
        self.cell = cell
        self.category = category
        self.chrom = chrom
        self.allele = allele
        self.start = start
        self.length = length
        self.event = event
        self.copies = copies
        self.homolog = homolog
        self.arm = arm
        self.ref_idx = ref_idx

def get_chrom_proportions(sequence):
    combined_chrom_len = dict(zip(sequence.keys(), map(lambda x: len(sequence[x][0]) + len(sequence[x][1]), sequence.keys())))
    total_len = sum(combined_chrom_len.values())
    combined_chrom_proportions = dict(zip(combined_chrom_len.keys(), [val/total_len for val in combined_chrom_len.values()]))
    return combined_chrom_proportions

# Creates initial genome from either from reference file, hg38 lengths, or with fixed chrom lengths and arm ratios
def init_diploid_genome(region_length, chrom_names, chrom_lens, arm_ratios):
    genome, regions = {}, {}
    static1, static2 = isinstance(chrom_lens, dict), isinstance(arm_ratios, dict)
    for chrom in chrom_names:
        if static1:
            chrom_len = chrom_lens[chrom]
        else:
            chrom_len = chrom_lens
        if static2:
            arm_ratio = arm_ratios[chrom]
        else:
            arm_ratio = arm_ratios

        num_regions = round(chrom_len / region_length)
        cent_idx = round(num_regions*arm_ratio)
        hap_profile = [i for i in range(num_regions)]
        hap_profile.insert(cent_idx, 'X')
        chrom_profile = [[[i for i in hap_profile]], [[i for i in hap_profile]]]
        genome[chrom] = chrom_profile
        regions[chrom] = num_regions
    return genome, regions



# Alters the genome of a cell based on the assignmened whole chrom events.
#def gen_whole_chrom_events(cell, chrom_names, chrom_arm_loc, chrom_event_rate, chrom_arm_rate, chrom_event_type, chrom_event_copy):
#    for chrom in chrom_names:
#        for allele in [0, 1]:
#            if np.random.binomial(1, chrom_event_rate) == 1:
#                if np.random.binomial(1, chrom_arm_rate) == 1:
#                    num_regions = len(cell.genome[chrom][allele])
#                    arm_region = chrom_arm_loc[chrom]
#                    while arm_region not in cell.genome[chrom][allele]:
#                        pass
#                    arm_idx = num_regions - cell.genome[chrom][allele][::-1].index(chrom_arm_loc[chrom])
#                    event_type = np.random.binomial(1, chrom_event_type)
#                    if np.random.binomial(1, 0.5) == 1:
#                        if event_type == 0:
#                            cell.genome[chrom][allele] = cell.genome[chrom][allele][:arm_idx]
#                    else:
#                        start_idx = arm_idx
#                        end_idx = num_regions
#                    event_type = np.random.binomial(1, chrom_event_type)
#                else:
#                    event_type = np.random.binomial(1, chrom_event_type)
#                    if event_type == 0:
#                        cell.genome[chrom][allele] = []
#                        cell.events.append(CNV(cell=cell, category=1, chrom=chrom, allele=allele, event=event_type))
#                    else:
#                        num_copies = np.random.geometric(chrom_event_copy)
#                        cell.genome[chrom][allele] = num_copies*cell.genome[chrom][allele]
#                        cell.events.append(CNV(cell=cell, category=1, chrom=chrom, allele=allele, event=event_type, copies=num_copies))

def gen_whole_chrom_event(cell, chrom_names, chrom_event_type):
    intact_chroms = []
    for chrom in chrom_names:
        for allele in [0, 1]:
            # Only consider chromosomes with both arms intact. If no X is present, can assume one of the arms was lost already.
            intact_chroms.extend([(chrom, allele, h) for h in range(len(cell.genome[chrom][allele])) if 'X' in cell.genome[chrom][allele][h]])
    if len(intact_chroms) == 0:
        print('No intact chromosomes remaining.')
        return
    
    chrom_idx = np.random.randint(len(intact_chroms))
    chrom, allele, homolog = intact_chroms[chrom_idx]

    event_type = np.random.binomial(1, chrom_event_type)
    if event_type == 0:
        del cell.genome[chrom][allele][homolog]
    else:
        cell.genome[chrom][allele].append([i for i in cell.genome[chrom][allele][homolog]])
    
    cell.events.append(CNV(cell=cell, category=1, chrom=chrom, allele=allele, homolog=homolog, event=event_type))

def gen_chrom_arm_event(cell, chrom_names, chrom_event_type):
    intact_arms = []
    for chrom in chrom_names:
        for allele in [0, 1]:
            for h in range(len(cell.genome[chrom][allele])):
                if len(cell.genome[chrom][allele][h]) > 0:
                    if 'X' in cell.genome[chrom][allele][h]:
                        split_idx = cell.genome[chrom][allele][h].index('X')
                        #if split_idx != 0:
                        intact_arms.append((chrom, allele, h, 'p'))
                        #if split_idx != len(cell.genome[chrom][allele][h]) - 1:
                        intact_arms.append((chrom, allele, h, 'q'))
                    else:
                        # Arm must have already been removed, but still consider the other one
                        intact_arms.append((chrom, allele, h, None))
    if len(intact_arms) == 0:
        print('No intact arms remaining.')
        return

    arm_idx = np.random.randint(len(intact_arms))
    chrom, allele, homolog, arm = intact_arms[arm_idx]
    if arm != None:
        split_idx = cell.genome[chrom][allele][homolog].index('X')
    
    event_type = np.random.binomial(1, chrom_event_type)
    if event_type == 0:
        if arm == 'p':
            cell.genome[chrom][allele][homolog] = cell.genome[chrom][allele][homolog][split_idx+1:]
        elif arm == 'q':
            cell.genome[chrom][allele][homolog] = cell.genome[chrom][allele][homolog][:split_idx]
        else:
            del cell.genome[chrom][allele][homolog]
    else:
        if arm == 'p':
            cell.genome[chrom][allele][homolog] = 2 * cell.genome[chrom][allele][homolog][:split_idx] + cell.genome[chrom][allele][homolog][split_idx:]
        elif arm == 'q':
            cell.genome[chrom][allele][homolog] = cell.genome[chrom][allele][homolog][:split_idx + 1] + 2 * cell.genome[chrom][allele][homolog][split_idx+1:]
        else:
            cell.genome[chrom][allele][homolog] = 2 * cell.genome[chrom][allele][homolog]
    
    cell.events.append(CNV(cell=cell, category=2, chrom=chrom, allele=allele, homolog=homolog, arm=arm, event=event_type))

def gen_focal_event(cell, chrom_names, length_mean, min_length, event_rate, copy_param):
    #Determine chrom and allele of event from those that haven't been lost.
    #CN_allele = np.random.binomial(1, 0.5)
    #chrom_lens = [len(cell.genome[chrom][CN_allele]) for chrom in chrom_names]
    intact_chroms = []
    for chrom in chrom_names:
            for allele in [0, 1]:
                #print(cell.name, chrom, allele, cell.genome[chrom][allele])
                intact_chroms.extend([(chrom, allele, h) for h in range(len(cell.genome[chrom][allele])) if len(cell.genome[chrom][allele][h]) > 0])
    chrom_proportions = [len(cell.genome[x[0]][x[1]][x[2]]) - 1 for x in intact_chroms]
    chrom_proportions = [x/sum(chrom_proportions) for x in chrom_proportions]
    chrom_idx = np.random.choice(list(range(len(intact_chroms))), p=chrom_proportions)
    CN_chrom, CN_allele, CN_homolog = intact_chroms[chrom_idx]

    #if sum(chrom_lens) == 0:
    #    print('Somehow all chroms of allele ' + str(CN_allele) + ' have been lost.')
    #chrom_proportions = [i / sum(chrom_lens) for i in chrom_lens]
    #CN_chrom = np.random.choice(chrom_names, p=chrom_proportions)

    #Determine properties of event
    num_regions = len(cell.genome[CN_chrom][CN_allele][CN_homolog])
    CN_size = max(min(round(np.random.exponential(length_mean)), num_regions), min_length)
    CN_type = np.random.binomial(1, event_rate)
    CN_copies = np.random.geometric(copy_param)

    #get starting location
    start_idx = np.random.randint(num_regions - CN_size + 1)
    ref_idx = cell.genome[CN_chrom][CN_allele][CN_homolog][start_idx]
    end_idx = start_idx + CN_size
    
    update_arm = False
    if 'X' in cell.genome[CN_chrom][CN_allele][CN_homolog]:
        arm_idx = cell.genome[CN_chrom][CN_allele][CN_homolog].index('X')
        if arm_idx >= start_idx and arm_idx <= end_idx:
            update_arm = True

    #update genome
    if CN_type == 1:
        gain_segment = CN_copies * cell.genome[CN_chrom][CN_allele][CN_homolog][start_idx:end_idx]
        if update_arm:
            gain_segment = list(filter(lambda x: x != 'X', gain_segment))
        cell.genome[CN_chrom][CN_allele][CN_homolog] = cell.genome[CN_chrom][CN_allele][CN_homolog][:end_idx] + gain_segment + cell.genome[CN_chrom][CN_allele][CN_homolog][end_idx:]
    else:
        cell.genome[CN_chrom][CN_allele][CN_homolog] = cell.genome[CN_chrom][CN_allele][CN_homolog][:start_idx] + cell.genome[CN_chrom][CN_allele][CN_homolog][end_idx:]
        if update_arm and CN_size != num_regions:
            cell.genome[CN_chrom][CN_allele][CN_homolog].insert(start_idx, 'X')
        if len(cell.genome[CN_chrom][CN_allele][CN_homolog]) == 0:
            del cell.genome[CN_chrom][CN_allele][CN_homolog]
        else:
            if cell.genome[CN_chrom][CN_allele][CN_homolog][0] == 'X':
                cell.genome[CN_chrom][CN_allele][CN_homolog].pop(0)
            if cell.genome[CN_chrom][CN_allele][CN_homolog][-1] == 'X':
                cell.genome[CN_chrom][CN_allele][CN_homolog].pop(-1)

    cell.events.append(CNV(cell=cell, category=0, chrom=CN_chrom, allele=CN_allele, start=start_idx, length=CN_size, event=CN_type, copies=CN_copies, ref_idx=ref_idx))

def mutate_genome(node, args, chrom_names):
    # Add WGD to founder node, if necessary
    if node.cell_type == 'founder':
        num_events = len(node.events)
        node.events = []
        if args['WGD']:
            for chrom in chrom_names:
                for allele in [0, 1]:
                    node.genome[chrom][allele].append([x for x in node.genome[chrom][allele][0]])

    # Add whole chrom events to founder and it's children, if necessary.
    if node.cell_type == 'founder' or node.parent.cell_type == 'founder' or node.cell_type == 'clone':
        if args['chrom_level_event']:
            if node.cell_type == 'clone':
                num_chrom_events = max(np.random.poisson(args['chrom_rate_clone']), 1) # we want to make sure theres atleast 1 event
            elif node.cell_type == 'founder':
                num_chrom_events = np.random.poisson(args['chrom_rate_founder']) # --> Changing implementation where this is a param specific to founder edges
            else:
                num_chrom_events = np.random.poisson(args['chrom_rate_super_clone'])
            for i in range(num_chrom_events):
                chrom_alt_type = np.random.binomial(1, args['chrom_arm_rate'])
                if chrom_alt_type == 1: #arm-level event
                    gen_chrom_arm_event(node, chrom_names, args['chrom_event_type'])
                else: #whome-chrom event
                    gen_whole_chrom_event(node, chrom_names, args['chrom_event_type'])
            #gen_whole_chrom_events(node, chrom_names, args['whole_chrom_rate'], args['whole_chrom_type'], args['whole_chrom_copy'])
    
    # Generates focal events. If above aneuploid tree, number of events precomputed.
    if node.cell_type == 'pseudonormal':
        num_events = len(node.events)
        node.events = []
    elif node.cell_type != 'founder': #foudner events already determined
        if args['placement_type'] == 0:
            num_events = np.random.poisson(args['placement_param'])
        if args['placement_type'] == 1:
            num_events = np.random.poisson(node.length)
        if args['placement_type'] == 2:
            num_events = int(args['placement_param'])
    #if node.is_root():
        #num_events = int(num_events * args['root_event_mult'])
    for i in range(num_events):
        gen_focal_event(node, chrom_names, args['cn_length_mean']/args['region_length'], max(round(args['min_cn_length']/args['region_length']), 1), args['cn_event_rate'], args['cn_copy_param'])

def genome_to_profile(node, chrom_names, num_regions, bins):
    # Collapsing genome in the form of a counter
    for chrom in chrom_names:
        for allele in [0, 1]:
            node.genome[chrom][allele] = [x for homolog in node.genome[chrom][allele] for x in homolog if x != 'X']
            node.genome[chrom][allele] = dict(Counter(node.genome[chrom][allele]))
            for i in set([i for i in range(num_regions[chrom])]) - set(node.genome[chrom][allele].keys()):
                node.genome[chrom][allele][i] = 0
    
    # Find the average number of copies over the regions in each bin
    node.profile = {}
    for chrom in chrom_names:
        node.profile[chrom] = [[], []]
        for allele in [0, 1]:
            for i in range(len(bins[chrom])-1):
                regions = list(range(bins[chrom][i], bins[chrom][i+1]))
                node.profile[chrom][allele].append(round(sum([node.genome[chrom][allele][j] for j in regions]) / len(regions)))

def evolve_tree(node, args, chrom_names, num_regions, bins=None):
    if not node.is_root():
        node.inheret()
    
    if not node.cell_type == 'normal':
        mutate_genome(node, args, chrom_names)
    
    if args['mode'] == 1 or args['mode'] == 2:
        with open(os.path.join(args['out_path'], node.name + '.pkl'), 'wb') as f:
            pickle.dump(node.genome, f)

    #if (args['mode'] == 0 or args['mode'] == 2) and bins != None:
    if node.is_leaf():
        genome_to_profile(node, chrom_names, num_regions, bins)
    else:
        with open(os.path.join(args['out_path'], node.name + '.pkl'), 'wb') as f:
            pickle.dump(node.genome, f)
            
    for child in node.children:
        evolve_tree(child, args, chrom_names, num_regions, bins=bins)

    del node.genome

def prepare_ancestral_profiles(tree, args, chrom_names, num_regions, bins=None):
    for node in tree.iter_preorder():
        if not node.is_leaf():
            with open(os.path.join(args['out_path'], node.name + '.pkl'), 'rb') as f:
                node.genome = pickle.load(f)
            genome_to_profile(node, chrom_names, num_regions, bins)
            del node.genome
            os.remove(os.path.join(args['out_path'], node.name + '.pkl'))



